from typing import Optional, Any, Union

from pydantic import BaseModel, validator

from unipipeline.utils.serializer_registry import serializer_registry, compressor_registry


class UniMessageCodec(BaseModel):
    compression: Optional[str]
    content_type: str

    @validator("content_type")
    def content_type_must_be_supported(cls, v: str) -> str:
        serializer_registry.assert_supports(v)
        return v

    @validator("compression")
    def compression_must_be_supported(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return None
        compressor_registry.assert_supports(v)
        return v

    def decompress(self, data: Union[bytes, str]) -> bytes:
        data_bytes: bytes
        if isinstance(data, str):
            data_bytes = bytes(data, encoding='utf-8')
        elif isinstance(data, bytes):
            data_bytes = data
        else:
            raise TypeError('invalid type')
        if self.compression is not None:
            return compressor_registry.loads(data_bytes, self.compression)  # type: ignore
        return data_bytes

    def loads(self, data: Union[bytes, str]) -> Any:
        data_str: str
        if isinstance(data, str):
            data_str = data
        elif isinstance(data, bytes):
            data_str = data.decode('utf-8')
        else:
            raise TypeError('invalid type')

        return serializer_registry.loads(data_str, self.content_type)  # type: ignore

    def dumps(self, data: Any) -> str:
        return serializer_registry.dumps(data, self.content_type)

    def compress(self, data: Union[bytes, str]) -> bytes:
        data_bytes: bytes
        if isinstance(data, str):
            data_bytes = bytes(data, encoding='utf-8')
        elif isinstance(data, bytes):
            data_bytes = data
        else:
            raise TypeError('invalid type')

        if self.compression is not None:
            return compressor_registry.dumps(data_bytes, self.compression)
        return data_bytes



import json
import logging
from logging import Logger
from typing import Set, List
from uuid import uuid4

from unipipeline.modules.uni_broker import UniBroker, UniBrokerConsumer
from unipipeline.modules.uni_message_meta import UniMessageMeta


class UniLogBroker(UniBroker):
    def start_consuming(self) -> None:
        self._logger.info(f'{self._logging_prefix} start consuming')

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        return 0

    def initialize(self, topics: Set[str]) -> None:
        self._logger.info(f'{self._logging_prefix} initialized')

    def mk_logger(self) -> Logger:
        return logging.getLogger(__name__)

    def mk_log_prefix(self) -> str:
        return f'{type(self).__name__} {self._uni_definition.name}::{uuid4()} :'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._logger = self.mk_logger()
        self._logging_prefix = self.mk_log_prefix()

    def connect(self) -> None:
        self._logger.info(f'{self._logging_prefix} connect')

    def close(self) -> None:
        self._logger.info(f'{self._logging_prefix} close')

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        self._logger.info(f'{self._logging_prefix} add consumer "{consumer.id}" to topic "{consumer.topic}" :: {consumer.group_id}')

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        for meta in meta_list:
            self._logger.info(f'{self._logging_prefix} publish {json.dumps(meta.dict())}')



from unipipeline import UniWorker

from example.messages.ender_message import EnderMessage


class EnderFristWorker(UniWorker):
    def handle_message(self, message: EnderMessage) -> None:
        raise NotImplementedError('method handle_message must be specified for class "EnderFristWorker"')



from typing import Callable, Any, Set, NamedTuple, List, TYPE_CHECKING, Generic, TypeVar, Type

from pydantic import ValidationError

from unipipeline.modules.uni_message_codec import UniMessageCodec
from unipipeline.modules.uni_definition import UniDynamicDefinition
from unipipeline.modules.uni_echo import UniEcho
from unipipeline.modules.uni_broker_definition import UniBrokerDefinition
from unipipeline.modules.uni_message_meta import UniMessageMeta

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator


class UniBrokerMessageManager:
    def reject(self) -> None:
        raise NotImplementedError(f'method reject must be specified for class "{type(self).__name__}"')

    def ack(self) -> None:
        raise NotImplementedError(f'method acknowledge must be specified for class "{type(self).__name__}"')


class UniBrokerConsumer(NamedTuple):
    topic: str
    id: str
    group_id: str
    message_handler: Callable[[UniMessageMeta, UniBrokerMessageManager], None]


TConf = TypeVar('TConf')


class UniBroker(Generic[TConf]):
    config_type: Type[TConf] = UniDynamicDefinition  # type: ignore

    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition) -> None:
        self._uni_definition = definition
        self._uni_mediator = mediator
        self._uni_echo = self._uni_mediator.echo.mk_child(f'broker[{self._uni_definition.name}]')

        try:
            self._uni_conf = self._uni_definition.configure_dynamic(self.config_type)  # type: ignore
        except ValidationError as e:
            self._uni_echo.exit_with_error(str(e))

    @property
    def config(self) -> TConf:
        return self._uni_conf

    def codec_serialize(self, meta: UniMessageMeta) -> bytes:
        meta_dumps = self.definition.codec.dumps(meta.dict())
        meta_compressed = self.definition.codec.compress(meta_dumps)
        return meta_compressed

    def codec_parse(self, content: bytes, codec: UniMessageCodec = None) -> UniMessageMeta:
        if codec is None:
            codec = self.definition.codec
        body_uncompressed = codec.decompress(content)
        body_json = codec.loads(body_uncompressed)
        return UniMessageMeta(**body_json)

    def connect(self) -> None:
        raise NotImplementedError(f'method connect must be implemented for {type(self).__name__}')

    def close(self) -> None:
        raise NotImplementedError(f'method close must be implemented for {type(self).__name__}')

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        raise NotImplementedError(f'method consume must be implemented for {type(self).__name__}')

    def stop_consuming(self) -> None:
        raise NotImplementedError(f'method stop_consuming must be implemented for {type(self).__name__}')

    def start_consuming(self) -> None:
        raise NotImplementedError(f'method start_consuming must be implemented for {type(self).__name__}')

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        raise NotImplementedError(f'method consume must be implemented for {type(self).__name__}')

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        raise NotImplementedError(f'method get_topic_size must be implemented for {type(self).__name__}')

    def initialize(self, topics: Set[str]) -> None:
        raise NotImplementedError(f'method initialize_topic must be implemented for {type(self).__name__}')

    @property
    def definition(self) -> UniBrokerDefinition:
        return self._uni_definition

    @property
    def echo(self) -> UniEcho:
        return self._uni_echo



from unipipeline.brokers.uni_amqp_broker import UniAmqpBroker, UniAmqpBrokerMessageManager, UniAmqpBrokerConfig
from unipipeline.brokers.uni_kafka_broker import UniKafkaBroker, UniKafkaBrokerMessageManager, UniKafkaBrokerConf
from unipipeline.brokers.uni_log_broker import UniLogBroker
from unipipeline.brokers.uni_memory_broker import UniMemoryBroker, UniMemoryBrokerMessageManager
from unipipeline.messages.uni_cron_message import UniCronMessage
from unipipeline.modules.uni import Uni
from unipipeline.modules.uni_broker import UniBrokerMessageManager, UniBroker, UniBrokerConsumer
from unipipeline.modules.uni_broker_definition import UniBrokerDefinition
from unipipeline.modules.uni_definition import UniDefinition
from unipipeline.modules.uni_external_definition import UniExternalDefinition
from unipipeline.modules.uni_message_codec import UniMessageCodec
from unipipeline.modules.uni_config import UniConfig, UniConfigError
from unipipeline.modules.uni_cron_job import UniCronJob
from unipipeline.modules.uni_cron_task_definition import UniCronTaskDefinition
from unipipeline.modules.uni_mediator import UniMediator
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta, UniMessageMetaErr, UniMessageMetaErrTopic
from unipipeline.modules.uni_message_definition import UniMessageDefinition
from unipipeline.modules.uni_module_definition import UniModuleDefinition
from unipipeline.modules.uni_service_definition import UniServiceDefinition
from unipipeline.modules.uni_waiting_definition import UniWaitingDefinition
from unipipeline.modules.uni_wating import UniWaiting
from unipipeline.modules.uni_worker import UniWorker, UniPayloadParsingError
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition
from unipipeline.utils.connection_pool import ConnectionObj, ConnectionRC, ConnectionManager, ConnectionPool, connection_pool
from unipipeline.utils.serializer_registry import SerializersRegistry, serializer_registry, compressor_registry, CONTENT_TYPE__APPLICATION_JSON, COMPRESSION__GZIP, \
    COMPRESSION__BZ2, COMPRESSION__LZMA

__all__ = (
    "Uni",
    "UniConfig",
    "UniConfigError",
    "UniMediator",
    "UniDefinition",
    "UniModuleDefinition",
    "UniServiceDefinition",
    "UniExternalDefinition",

    "SerializersRegistry",
    "serializer_registry",
    "compressor_registry",
    "CONTENT_TYPE__APPLICATION_JSON",
    "COMPRESSION__GZIP",
    "COMPRESSION__BZ2",
    "COMPRESSION__LZMA",

    "ConnectionObj",
    "ConnectionRC",
    "ConnectionManager",
    "ConnectionPool",
    "connection_pool",

    # cron
    "UniCronJob",
    "UniCronTaskDefinition",

    # broker
    "UniBrokerMessageManager",
    "UniBroker",
    "UniBrokerConsumer",
    "UniBrokerDefinition",

    "UniAmqpBroker",
    "UniAmqpBrokerConfig",
    "UniAmqpBrokerMessageManager",

    "UniLogBroker",

    "UniMemoryBroker",
    "UniMemoryBrokerMessageManager",

    "UniKafkaBroker",
    "UniKafkaBrokerConf",
    "UniKafkaBrokerMessageManager",

    # message
    "UniMessage",
    "UniMessageDefinition",
    "UniMessageMeta",
    "UniMessageMetaErr",
    "UniMessageMetaErrTopic",
    "UniCronMessage",

    # worker
    "UniPayloadParsingError",
    "UniWorker",
    "UniWorkerDefinition",

    # waiting
    "UniWaitingDefinition",
    "UniWaiting",
)



import time
from time import sleep
from typing import Optional, TypeVar, Set, List, NamedTuple, Callable
from urllib.parse import urlparse

from pika import ConnectionParameters, PlainCredentials, BlockingConnection, BasicProperties, spec  # type: ignore
from pika.adapters.blocking_connection import BlockingChannel  # type: ignore
from pika.exceptions import AMQPConnectionError, AMQPError, ConnectionClosedByBroker  # type: ignore

from unipipeline.modules.uni_broker import UniBroker, UniBrokerMessageManager, UniBrokerConsumer
from unipipeline.modules.uni_definition import UniDynamicDefinition
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_codec import UniMessageCodec
from unipipeline.modules.uni_message_meta import UniMessageMeta

BASIC_PROPERTIES__HEADER__COMPRESSION_KEY = 'compression'


TMessage = TypeVar('TMessage', bound=UniMessage)


class UniAmqpBrokerMessageManager(UniBrokerMessageManager):

    def __init__(self, channel: BlockingChannel, method_frame: spec.Basic.Deliver) -> None:
        self._channel = channel
        self._method_frame = method_frame
        self._acknowledged = False

    def reject(self) -> None:
        self._channel.basic_reject(delivery_tag=self._method_frame.delivery_tag, requeue=True)

    def ack(self) -> None:
        if self._acknowledged:
            return
        self._acknowledged = True
        self._channel.basic_ack(delivery_tag=self._method_frame.delivery_tag)


class UniAmqpBrokerConfig(UniDynamicDefinition):
    exchange_name: str = "communication"
    heartbeat: int = 600
    blocked_connection_timeout: int = 300
    prefetch: int = 1
    retry_max_count: int = 100
    retry_delay_s: int = 3
    socket_timeout: int = 300
    stack_timeout: int = 300
    exchange_type: str = "direct"
    durable: bool = True
    auto_delete: bool = False
    passive: bool = False
    is_persistent: bool = True


class UniAmqpBrokerConsumer(NamedTuple):
    queue: str
    on_message_callback: Callable[[BlockingChannel, spec.Basic.Deliver, BasicProperties, bytes], None]
    consumer_tag: str


class UniAmqpBroker(UniBroker[UniAmqpBrokerConfig]):
    config_type = UniAmqpBrokerConfig

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        ch = self._get_channel()
        res = ch.queue_declare(
            queue=topic,
            passive=True
        )
        return res.method.message_count

    @classmethod
    def get_connection_uri(cls) -> str:
        raise NotImplementedError(f"cls method get_connection_uri must be implemented for class '{cls.__name__}'")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        broker_url = self.get_connection_uri()

        url_params_pr = urlparse(url=broker_url)

        self._params = ConnectionParameters(
            heartbeat=self.config.heartbeat,
            blocked_connection_timeout=self.config.blocked_connection_timeout,
            socket_timeout=self.config.socket_timeout,
            stack_timeout=self.config.stack_timeout,
            retry_delay=self.definition.retry_delay_s,
            host=url_params_pr.hostname,
            port=url_params_pr.port,
            credentials=PlainCredentials(url_params_pr.username, url_params_pr.password, erase_on_connect=False),
        )

        self._consumers: List[UniAmqpBrokerConsumer] = list()

        self._connection: Optional[BlockingConnection] = None
        self._channel: Optional[BlockingChannel] = None

        self._consuming_started = False
        self._in_processing = False
        self._interrupted = False

    def initialize(self, topics: Set[str]) -> None:
        ch = self._get_channel()
        ch.exchange_declare(
            exchange=self.config.exchange_name,
            exchange_type=self.config.exchange_type,
            passive=self.config.passive,
            durable=self.config.durable,
            auto_delete=self.config.auto_delete,
        )

        ch.basic_qos(prefetch_count=self.config.prefetch)

        for topic in topics:
            ch.queue_declare(
                queue=topic,
                durable=self.config.durable,
                auto_delete=self.config.auto_delete,
                passive=False,
            )

            ch.queue_bind(queue=topic, exchange=self.config.exchange_name, routing_key=topic)

    def stop_consuming(self) -> None:
        self._end_consuming()

    def _end_consuming(self):
        if not self._consuming_started:
            return
        self._interrupted = True
        if not self._in_processing:
            self._get_channel().stop_consuming()
            self.close()
            self._consuming_started = False
            self.echo.log_info('consumption stopped')

    def connect(self) -> None:
        if self._connection is not None:
            if self._connection.is_closed:
                self._connection = None
            else:
                return

        if self._channel is not None:
            if self._channel.is_closed:
                self._channel = None
            else:
                return

        try:
            self._connection = BlockingConnection(self._params)
            self._channel = self._connection.channel()
        except (AMQPError, AMQPConnectionError) as e:
            raise ConnectionError(str(e))

        self.echo.log_info('connected')

    def close(self) -> None:
        try:
            if self._channel is not None and not self._channel.is_closed:
                self._channel.close()
        except AMQPError:
            pass

        try:
            if self._connection is not None and not self._connection.is_closed:
                self._connection.close()
        except AMQPError:
            pass

        self._connection = None
        self._channel = None

    def _get_channel(self) -> BlockingChannel:
        self.connect()
        assert self._channel is not None
        return self._channel

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        echo = self.echo.mk_child(f'topic[{consumer.topic}]')
        if self._consuming_started:
            echo.log_error(f'you cannot add consumer dynamically :: tag="{consumer.id}" group_id={consumer.group_id}')
            exit(1)

        def consumer_wrapper(channel: BlockingChannel, method_frame: spec.Basic.Deliver, properties: BasicProperties, body: bytes) -> None:
            self._in_processing = True
            meta = self.codec_parse(body, UniMessageCodec(
                compression=properties.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
                content_type=properties.content_type
            ))
            manager = UniAmqpBrokerMessageManager(channel, method_frame)
            consumer.message_handler(meta, manager)
            self._in_processing = False
            if self._interrupted:
                self._end_consuming()

        self._consumers.append(UniAmqpBrokerConsumer(
            queue=consumer.topic,
            on_message_callback=consumer_wrapper,
            consumer_tag=consumer.id,
        ))

        echo.log_info(f'added consumer :: tag="{consumer.id}" group_id={consumer.group_id}')

    def start_consuming(self) -> None:
        echo = self.echo.mk_child('consuming')
        if len(self._consumers) == 0:
            echo.log_warning('has no consumers to start consuming')
            return
        if self._consuming_started:
            echo.log_warning('consuming has already started. ignored')
            return
        self._consuming_started = True
        self._interrupted = False
        self._in_processing = False

        retry_counter = 0
        retry_threshold_s = self.config.retry_delay_s * (self.config.retry_max_count + 1)
        while True:
            start = time.time()
            try:
                ch = self._get_channel()
                for c in self._consumers:
                    ch.basic_consume(queue=c.queue, on_message_callback=c.on_message_callback, consumer_tag=c.consumer_tag)
                echo.log_info(f'consumers count is {len(self._consumers)}')
                ch.start_consuming()  # blocking operation
            except (ConnectionClosedByBroker, ConnectionError) as e:
                end = time.time()
                echo.log_error(f'connection closed {e}')
                if int(end - start) >= retry_threshold_s:
                    retry_counter = 0
                if retry_counter >= self.config.retry_max_count:
                    raise ConnectionError()
                retry_counter += 1
                sleep(self.config.retry_delay_s)

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        for meta in meta_list: # TODO: package sending
            # TODO: retry
            self._get_channel().basic_publish(
                exchange=self.config.exchange_name,
                routing_key=topic,
                body=self.codec_serialize(meta),
                properties=BasicProperties(
                    content_type=self.definition.codec.content_type,
                    content_encoding='utf-8',
                    delivery_mode=2 if self.config.is_persistent else 0,
                    headers={
                        BASIC_PROPERTIES__HEADER__COMPRESSION_KEY: self.definition.codec.compression
                    }
                )
            )



from typing import Dict, Any, Set, Iterator, Tuple
from uuid import uuid4


COMMON_PROP = "__default__"


class ParseDefinitionError(Exception):
    pass


def parse_definition(
    conf_name: str,
    definitions: Dict[str, Any],
    defaults: Dict[str, Any],
    required_keys: Set[str]
) -> Iterator[Tuple[str, Dict[str, Any], Dict[str, Any]]]:
    if not isinstance(definitions, dict):
        raise ParseDefinitionError(f'definition of {conf_name} has invalid type. must be dict')

    common = definitions.get(COMMON_PROP, dict())

    for name, raw_definition in definitions.items():
        other_props: Dict[str, Any] = dict()

        name = str(name)
        if name.startswith("_"):
            if name == COMMON_PROP:
                continue
            else:
                raise ValueError(f"key '{name}' is not acceptable")

        if not isinstance(raw_definition, dict):
            raise ParseDefinitionError(f'definition of {conf_name}->{name} has invalid type. must be dict')

        result_definition = dict(defaults)
        combined_definition = dict(common)
        combined_definition.update(raw_definition)

        for k, v in combined_definition.items():
            if k in defaults:
                result_definition[k] = v
                vd = defaults.get(k, None)
                if vd is not None and type(vd) != type(v):
                    raise ParseDefinitionError(f'definition of {conf_name}->{name} has invalid key "{k}" type')
            elif k in required_keys:
                result_definition[k] = v
            else:
                other_props[k] = v

        for rk in required_keys:
            if rk not in result_definition:
                raise ParseDefinitionError(f'definition of {conf_name}->{name} has no required prop "{rk}"')

        result_definition["name"] = name
        result_definition["id"] = uuid4()

        yield name, result_definition, other_props



from typing import TypeVar, Optional
from uuid import UUID

from unipipeline.modules.uni_definition import UniDefinition
from unipipeline.modules.uni_message_codec import UniMessageCodec
from unipipeline.modules.uni_module_definition import UniModuleDefinition

TContent = TypeVar('TContent')


class UniBrokerDefinition(UniDefinition):
    id: UUID
    type: UniModuleDefinition

    retry_max_count: int
    retry_delay_s: int

    external: Optional[str]

    codec: UniMessageCodec

    @property
    def marked_as_external(self) -> bool:
        return self.external is not None



from typing import Any, Dict, TypeVar, Type

from pydantic import BaseModel


class UniDynamicDefinition(BaseModel):
    pass


T = TypeVar('T', bound=UniDynamicDefinition)


class UniDefinition(BaseModel):
    name: str

    dynamic_props_: Dict[str, Any]

    def configure_dynamic(self, dynamic_type: Type[T]) -> T:
        if not issubclass(dynamic_type, UniDynamicDefinition):
            raise TypeError(f'dynamic prop has invalid type "{dynamic_type.__name__}". must be subclass UniDynamicDefinition')
        return dynamic_type(**self.dynamic_props_)



import bz2
import json
import lzma
import zlib
from datetime import datetime
from enum import Enum
from typing import Dict, Any, Callable, Set, Type, TypeVar, Generic
from uuid import UUID

CONTENT_TYPE__APPLICATION_JSON = 'application/json'
# CONTENT_TYPE__APPLICATION_DATA = 'application/data'

COMPRESSION__GZIP = 'application/x-gzip'
COMPRESSION__BZ2 = 'application/x-bz2'
COMPRESSION__LZMA = 'application/x-lzma'


TContentEncoding = str
TContentType = str


T = TypeVar('T')


def ensure_type(name: str, value: T, constructor: Type[T]) -> T:
    if not isinstance(value, constructor):
        raise TypeError(f'{name} has invalid type. must be {constructor.__name__}, {type(value).__name__} was given')
    return value


def prepare_content_type(content_type: TContentType) -> TContentType:
    ensure_type('content_type', content_type, TContentType)
    return content_type.lower()


TSerializedData = TypeVar("TSerializedData")
TParsedData = TypeVar("TParsedData")


class SerializersRegistry(Generic[TParsedData, TSerializedData]):
    def __init__(self) -> None:
        self._encoders: Dict[TContentType, Callable[[TParsedData], TSerializedData]] = dict()
        self._decoders: Dict[TContentType, Callable[[TSerializedData], TParsedData]] = dict()
        self._content_types: Set[TContentType] = set()

    def register(self, content_type: TContentType, encoder: Callable[[TParsedData], TSerializedData], decoder: Callable[[TSerializedData], TParsedData]) -> None:
        content_type = prepare_content_type(content_type)
        self._encoders[content_type] = encoder
        self._decoders[content_type] = decoder
        self._content_types.add(content_type)

    def supports(self, content_type: TContentType) -> bool:
        content_type = prepare_content_type(content_type)
        return content_type in self._content_types

    def assert_supports(self, content_type: TContentType) -> None:
        content_type = prepare_content_type(content_type)
        if content_type not in self._content_types:
            raise TypeError(f'content_type "{content_type}" is not supported')

    def dumps(self, data: TParsedData, content_type: TContentType) -> TSerializedData:
        content_type = prepare_content_type(content_type)
        encoder = self._encoders[content_type]
        return encoder(data)

    def loads(self, data: TSerializedData, content_type: TContentType) -> TParsedData:
        content_type = prepare_content_type(content_type)
        decoder = self._decoders[content_type]
        return decoder(data)


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, bytes):
            return obj.hex()
        if isinstance(obj, Enum):
            return obj.value
        return json.JSONEncoder.default(self, obj)


def json_dumps(data: Any) -> str:
    return ComplexEncoder().encode(data)


serializer_registry = SerializersRegistry[Any, str]()
serializer_registry.register(CONTENT_TYPE__APPLICATION_JSON, json_dumps, json.loads)

compressor_registry = SerializersRegistry[bytes, bytes]()
compressor_registry.register(COMPRESSION__GZIP, zlib.compress, zlib.decompress)
compressor_registry.register(COMPRESSION__BZ2, bz2.compress, bz2.decompress)
compressor_registry.register(COMPRESSION__LZMA, lzma.compress, lzma.decompress)
